﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using IdeaBlade.Core;
using IdeaBlade.Core.Composition;
using IdeaBlade.EntityModel;
using SimpleNorthwindModel;

namespace SimpleComboDemo {
  /// <summary>
  /// MainPageViewModel to use as production ViewModel for MainPage view.
  /// </summary>
  /// <remarks>
  /// Does NOT exhibit best practices for a ViewModel.
  /// The direct awareness of and involvement with an EntityManager
  /// is strong indicator of poor design.
  /// It's a demo of a ViewModel
  /// </remarks>
  public class MainPageViewModel : INotifyPropertyChanged {

    protected NorthwindManager EntityManager;

    /// <summary>Design Time constructor.</summary>
    public MainPageViewModel() {
      Messages = new ObservableCollection<string>();
      Employees = new ObservableCollection<Employee>();
    }

    /// <summary>Production constructor.</summary>
    /// <remarks>Real version would be injected with stuff.</remarks>
    public MainPageViewModel(object irrelevantParameterAtTheMoment) : this() {
      EntityManager = new NorthwindManager();
      GetPotentialManagers();
    }

    #region PotentialManagers

    private async void GetPotentialManagers() {
      WriteMessage("Getting Potential Managers");
      // Everybody has managerial potential :-)
      try{
        var results = await EntityManager.Employees.ExecuteAsync();
        GotPotentialManagers(results);
      }catch (Exception e) {
          WriteMessage("!!! Potential Managers query failed on server: " + e.Message);
        }
    }

    private void GotPotentialManagers(IEnumerable<Employee> results) {

        var mgrs = results.OrderBy(mgr => mgr.FirstName) // results are sorted on client, not server
          .ToList();

        mgrs.Insert(0, EntityManager.GetNullEntity<Employee>()); // represents "No Manager"
        PotentialManagers = mgrs;
        WriteMessage("Got Potential Managers");

        GetEmployees();
    }

    /// <summary>
    /// Get the list of employees who could be managers for some employee.
    /// This list is constant for the application.
    /// Compare to <see cref="CandidateManagers"/>
    /// </summary>
    public IList<Employee> PotentialManagers {
      get { return _potentialManagers; }
      protected set { _potentialManagers = value; RaisePropertyChanged("PotentialManagers"); }
    }
    private IList<Employee> _potentialManagers;

    #endregion

    #region Employees / CurrentEmployee

    private async void GetEmployees() {
      WriteMessage("Getting Employees");
      try {
        var results = await EntityManager.Employees.ExecuteAsync();
        GotEmployees(results);
      } catch (Exception e) {
        WriteMessage("!!! Employees query failed on server: " + e.Message);
      }
    }

    private void GotEmployees(IEnumerable<Employee> results) {
      results.ForEach(Employees.Add);
      WriteMessage("Got Employees");
      CurrentEmployee = Employees.FirstOrDefault();
      IsReady = true;
    }

    public ObservableCollection<Employee> Employees { get; private set; }

    public Employee CurrentEmployee {
      get { return _currentEmployee; }
      set {
        if (_currentEmployee == value) return;
        _currentEmployee = value;
        SetCandidateManagersForCurrentEmployee();
        RaisePropertyChanged("CurrentEmployee");
        WriteCurrentEmployeeMessage();
      }
    }
    private Employee _currentEmployee;

    public bool IsReady {
      get { return _isReady; }
      set { _isReady = value; RaisePropertyChanged("IsReady"); }
    }
    private bool _isReady;
    #endregion

    #region CandidateManagers

    ///<summary>
    /// Get those folks who could be the current employee's manager
    /// </summary>
    /// Compare to <see cref="PotentialManagers"/>
    public IList<Employee> CandidateManagers { get; private set; }


    /// <summary>
    /// Sets the list of employees who are allowed to be managers of the current employee.
    /// This list changes from employee to employee.
    /// See <see cref="CandidateManagers"/>
    /// </summary>
    private void SetCandidateManagersForCurrentEmployee() {

      // Tricky, tricky !
      // Must unbind CurrentEmployee (or find another way to disable binding)
      // while swapping the CandidateManagers list or else
      // binding will set the CurrentEmployee's manager to null.

      // None of this would be necessary if the list did not change between employees
      // as when (if) we bound to PotentialManagers instead of CandidateManagers.

      var curEmp = _currentEmployee;
      _currentEmployee = null;
      RaisePropertyChanged("CurrentEmployee");

      SetCandidateManagersForEmployee(curEmp);
      RaisePropertyChanged("CandidateManagers");

      // Now we can restore the CurrentEmployee
      // We'll let caller re-raise CurrentEmployee
      _currentEmployee = curEmp; // restore current employee
    }

    /// <summary>
    /// Sets the list of employees who are allowed to be managers of the employee.
    /// This list changes from employee to employee.
    /// See <see cref="SetCandidateManagersForCurrentEmployee"/>
    /// </summary>
    /// <remarks>
    /// Determines candidate managers for the employee
    /// According to "Bob", you can't manage yourself so remove self but ...
    /// ... always keep whatever manager you currently have in the list
    /// because not trying to change or clear manager this way
    /// </remarks>
    private void SetCandidateManagersForEmployee(Employee emp) {

      // Comprehension syntax
      var q = from mgr in (_potentialManagers ?? new List<Employee>())
              where mgr != emp || mgr == emp.Manager
              select mgr;
      CandidateManagers = q.ToList();

      // Lambda syntax
      // _candidateManagers =  (_potentialManagers ?? new List<Employee>())
      //    .Where(mgr => mgr != emp || mgr == emp.Manager)
      //    .ToList();     

      RaisePropertyChanged("CandidateManagers");
    }

    #endregion

    #region Actions

    public void GoBack() {
      var ix = Employees.IndexOf(CurrentEmployee) - 1;
      if (ix > -1) CurrentEmployee = Employees[ix];
    }

    public void GoNext() {
      var ix = Employees.IndexOf(CurrentEmployee) + 1;
      if (ix < Employees.Count) CurrentEmployee = Employees[ix];
    }

    public async void Save() {
      WriteMessage("Saving Employee(s)");
      try {
        await EntityManager.SaveChangesAsync();
      } catch (Exception e) {
        WriteMessage("!!! Save failed on server: " + e.Message);
      }
    }

    public void Undo() {
      CurrentEmployee.EntityAspect.RejectChanges();
    }

    #endregion

    #region Log Messages

    public ObservableCollection<string> Messages {
      get { return _messages; }
      private set { _messages = value; RaisePropertyChanged("Messages"); }
    }
    private ObservableCollection<string> _messages;

    private int _messageCounter = 1;
    private void WriteMessage(string msg) {
      Messages.Insert(0, string.Format("{0:d4}: {1}", _messageCounter++, msg)); // insert at top
    }

    private void WriteCurrentEmployeeMessage() {
      WriteMessage(
        String.Format("Current {0} employee: {1} whose manager is {2}",
                      _currentEmployee.EntityAspect.EntityState,
                      _currentEmployee.FullName, _currentEmployee.Manager.FullName));
    }

    #endregion

    #region INotifyPropertyChanged Members

    public event PropertyChangedEventHandler PropertyChanged;

    private void RaisePropertyChanged(string propertyName) {
      if (null == PropertyChanged) return;
      PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
    }

    #endregion

  }
}
