﻿Imports System.Collections.ObjectModel
Imports System.ComponentModel
Imports IdeaBlade.Core
Imports IdeaBlade.Core.Composition
Imports IdeaBlade.EntityModel
Imports SimpleNorthwindModel


''' <summary>
''' MainPageViewModel to use as production ViewModel for MainPage view.
''' </summary>
''' <remarks>
''' Does NOT exhibit best practices for a ViewModel.
''' The direct awareness of and involvement with an EntityManager
''' is strong indicator of poor design.
''' It's a demo of a ViewModel
''' </remarks>
Public Class MainPageViewModel
  Implements INotifyPropertyChanged

  Protected EntityManager As NorthwindManager

  ''' <summary>Design Time constructor.</summary>
  Public Sub New()
    Messages = New ObservableCollection(Of String)()
    Employees = New ObservableCollection(Of Employee)()
  End Sub

  ''' <summary>Production constructor.</summary>
  ''' <remarks>Real version would be injected with stuff.</remarks>
  Public Sub New(ByVal irrelevantParameterAtTheMoment As Object)
    Me.New()
    EntityManager = New NorthwindManager()
    GetPotentialManagers()
  End Sub

#Region "PotentialManagers"

  Private Async Sub GetPotentialManagers()
    WriteMessage("Getting Potential Managers")
    ' Everybody has managerial potential :-)
    Try
      Dim results = Await EntityManager.Employees.ExecuteAsync()
      GotPotentialManagers(results)
    Catch ex As Exception
      WriteMessage("!!! Potential Managers query failed on server: " & ex.Message)
    End Try
  End Sub

  Private Sub GotPotentialManagers(ByVal results As IEnumerable(Of Employee))
    Dim mgrs = results.OrderBy(Function(mgr) mgr.FirstName).ToList() ' results are sorted on client, not server

    mgrs.Insert(0, EntityManager.GetNullEntity(Of Employee)()) ' represents "No Manager"
    PotentialManagers = mgrs
    WriteMessage("Got Potential Managers")

    GetEmployees()
  End Sub

  ''' <summary>
  ''' Get the list of employees who could be managers for some employee.
  ''' This list is constant for the application.
  ''' Compare to <see cref="CandidateManagers"/>
  ''' </summary>
  Public Property PotentialManagers() As IList(Of Employee)
    Get
      Return _potentialManagers
    End Get
    Protected Set(ByVal value As IList(Of Employee))
      _potentialManagers = value
      RaisePropertyChanged("PotentialManagers")
    End Set
  End Property
  Private _potentialManagers As IList(Of Employee)

#End Region

#Region "Employees / CurrentEmployee"

  Private Async Sub GetEmployees()
    WriteMessage("Getting Employees")
    Try
      Dim results = Await EntityManager.Employees.ExecuteAsync()
      GotEmployees(results)
    Catch ex As Exception
      WriteMessage("!!! Employees query failed on server: " & ex.Message)
    End Try
  End Sub

  Private Sub GotEmployees(ByVal results As IEnumerable(Of Employee))
    results.ForEach(Sub(c) Employees.Add(c))
    WriteMessage("Got Employees")
    CurrentEmployee = Employees.FirstOrDefault()
    IsReady = True
  End Sub

  Private privateEmployees As ObservableCollection(Of Employee)
  Public Property Employees() As ObservableCollection(Of Employee)
    Get
      Return privateEmployees
    End Get
    Private Set(ByVal value As ObservableCollection(Of Employee))
      privateEmployees = value
    End Set
  End Property

  Public Property CurrentEmployee() As Employee
    Get
      Return _currentEmployee
    End Get
    Set(ByVal value As Employee)
      If _currentEmployee Is value Then
        Return
      End If
      _currentEmployee = value
      SetCandidateManagersForCurrentEmployee()
      RaisePropertyChanged("CurrentEmployee")
      WriteCurrentEmployeeMessage()
    End Set
  End Property
  Private _currentEmployee As Employee

  Public Property IsReady() As Boolean
    Get
      Return _isReady
    End Get
    Set(ByVal value As Boolean)
      _isReady = value
      RaisePropertyChanged("IsReady")
    End Set
  End Property
  Private _isReady As Boolean
#End Region

#Region "CandidateManagers"

  '''<summary>
  ''' Get those folks who could be the current employee's manager
  ''' </summary>
  ''' Compare to <see cref="PotentialManagers"/>
  Private privateCandidateManagers As IList(Of Employee)
  Public Property CandidateManagers() As IList(Of Employee)
    Get
      Return privateCandidateManagers
    End Get
    Private Set(ByVal value As IList(Of Employee))
      privateCandidateManagers = value
    End Set
  End Property


  ''' <summary>
  ''' Sets the list of employees who are allowed to be managers of the current employee.
  ''' This list changes from employee to employee.
  ''' See <see cref="CandidateManagers"/>
  ''' </summary>
  Private Sub SetCandidateManagersForCurrentEmployee()

    ' Tricky, tricky !
    ' Must unbind CurrentEmployee (or find another way to disable binding)
    ' while swapping the CandidateManagers list or else
    ' binding will set the CurrentEmployee's manager to null.

    ' None of this would be necessary if the list did not change between employees
    ' as when (if) we bound to PotentialManagers instead of CandidateManagers.

    Dim curEmp = _currentEmployee
    _currentEmployee = Nothing
    RaisePropertyChanged("CurrentEmployee")

    SetCandidateManagersForEmployee(curEmp)
    RaisePropertyChanged("CandidateManagers")

    ' Now we can restore the CurrentEmployee
    ' We'll let caller re-raise CurrentEmployee
    _currentEmployee = curEmp ' restore current employee
  End Sub

  ''' <summary>
  ''' Sets the list of employees who are allowed to be managers of the employee.
  ''' This list changes from employee to employee.
  ''' See <see cref="SetCandidateManagersForCurrentEmployee"/>
  ''' </summary>
  ''' <remarks>
  ''' Determines candidate managers for the employee
  ''' According to "Bob", you can't manage yourself so remove self but ...
  ''' ... always keep whatever manager you currently have in the list
  ''' because not trying to change or clear manager this way
  ''' </remarks>
  Private Sub SetCandidateManagersForEmployee(ByVal emp As Employee)

    ' Comprehension syntax
    Dim q = From mgr In (If(_potentialManagers, New List(Of Employee)()))
            Where mgr IsNot emp OrElse mgr Is emp.Manager
            Select mgr
    CandidateManagers = q.ToList()

    ' Lambda syntax
    ' _candidateManagers =  (_potentialManagers ?? new List<Employee>())
    '    .Where(mgr => mgr != emp || mgr == emp.Manager)
    '    .ToList();     

    RaisePropertyChanged("CandidateManagers")
  End Sub

#End Region

#Region "Actions"

  Public Sub GoBack()
    Dim ix = Employees.IndexOf(CurrentEmployee) - 1
    If ix > -1 Then
      CurrentEmployee = Employees(ix)
    End If
  End Sub

  Public Sub GoNext()
    Dim ix = Employees.IndexOf(CurrentEmployee) + 1
    If ix < Employees.Count Then
      CurrentEmployee = Employees(ix)
    End If
  End Sub

  Public Async Sub Save()
    WriteMessage("Saving Employee(s)")
    Try
      Await EntityManager.SaveChangesAsync()
    Catch ex As Exception
      WriteMessage("!!! Save failed on server: " & ex.Message)
    End Try
  End Sub

  Public Sub Undo()
    CurrentEmployee.EntityAspect.RejectChanges()
  End Sub

#End Region

#Region "Log Messages"

  Public Property Messages() As ObservableCollection(Of String)
    Get
      Return _messages
    End Get
    Private Set(ByVal value As ObservableCollection(Of String))
      _messages = value
      RaisePropertyChanged("Messages")
    End Set
  End Property
  Private _messages As ObservableCollection(Of String)

  Private _messageCounter As Integer = 1
  Private Sub WriteMessage(ByVal msg As String)
    Messages.Insert(0, String.Format("{0:d4}: {1}", _messageCounter, msg))
    _messageCounter += 1
  End Sub

  Private Sub WriteCurrentEmployeeMessage()
    WriteMessage(String.Format("Current {0} employee: {1} whose manager is {2}", _currentEmployee.EntityAspect.EntityState, _currentEmployee.FullName, _currentEmployee.Manager.FullName))
  End Sub

#End Region

#Region "INotifyPropertyChanged Members"

  Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged

  Private Sub RaisePropertyChanged(ByVal propertyName As String)
    'If Nothing Is PropertyChanged Then
    ' Return
    'End If
    RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
  End Sub

#End Region

End Class

